
/******************************************************************************
* MODULE     : base.hpp
* DESCRIPTION: Basic inclusions, definitions and routines
* COPYRIGHT  : (C) 2003  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_BASE_HPP
#define __MML_BASE_HPP
#include "config.hpp"

#ifndef MMX_GC
#define MMX_GC
#ifdef MML_SUPPORTS_GC
#ifdef MML_OLD_CPP
#include <gc.h>
inline void* operator new (std::size_t sz) { return GC_MALLOC (sz); }
inline void* operator new[] (std::size_t sz) { return GC_MALLOC (sz); }
inline void operator delete (void* dummy) { (void) dummy; }
inline void operator delete[] (void* dummy ) { (void) dummy; }
#else
#include <gc/new_gc_alloc.h>
#include <new>
inline void* operator new (std::size_t sz) throw (std::bad_alloc) {
  return GC_MALLOC (sz); }
inline void* operator new[] (std::size_t sz) throw (std::bad_alloc) {
  return GC_MALLOC (sz); }
inline void operator delete (void*) throw() {}
inline void operator delete[] (void*) throw() {}
inline void* operator new (std::size_t sz, const std::nothrow_t& d) throw() {
  (void) d; return GC_MALLOC (sz); }
inline void* operator new[] (std::size_t sz, const std::nothrow_t& d) throw() {
  (void) d; return GC_MALLOC (sz); }
inline void operator delete (void*, const std::nothrow_t& d) throw() {
  (void) d; }
inline void operator delete[] (void*, const std::nothrow_t& d) throw() {
  (void) d; }
#endif
#endif
#endif

#include <stdlib.h>
#include <iostream>
#include <cassert>
using std::istream;
using std::ostream;
using std::cin;
using std::cout;
using std::cerr;
namespace mml {

void mml_init ();
volatile void fatal_error (char* message, char* routine, char* file);
typedef unsigned int nat;

struct rep_struct {
  int ref_count;
  inline rep_struct (): ref_count (1) {}
  virtual inline ~rep_struct () {}
};

#define INC_COUNT(x) { if ((x)!=NULL) (x)->ref_count++; }
#define DEC_COUNT(x) { if ((x)!=NULL && 0==--((x)->ref_count)) delete (x); }

/******************************************************************************
* Indirect data structures prototypes
******************************************************************************/
  
#define INDIRECT_PROTO(T,R)			\
protected:					\
  R* rep;					\
public:						\
  inline T (R* rep2);				\
  inline T (R* rep2, bool with_inc);		\
  inline T (const T& x);			\
  inline ~T ();					\
  inline T& operator= (const T& x);		\
  inline R* operator-> () const;
  
#define INDIRECT_STATIC_PROTO(T,R)		\
protected:					\
  R* rep;					\
public:						\
  T (const T& x);				\
  ~T ();					\
  T& operator= (const T& x);			\
  R* operator-> () const;

#define INDIRECT_PROTO_1(T,R,X)			\
protected:					\
  R<X >* rep;					\
public:						\
  T (R<X >* rep2);				\
  T (const T& x);				\
  ~T ();					\
  T& operator= (const T& x);			\
  R<X >* operator-> () const;

#define INDIRECT_PROTO_2(T,R,X,Y)		\
protected:					\
  R<X,Y >* rep;					\
public:						\
  T (R<X,Y >* rep2);				\
  T (const T& x);				\
  ~T ();					\
  T& operator= (const T& x);			\
  R<X,Y >* operator-> () const;

#define INDIRECT_PROTO_3(T,R,X,Y,Z)		\
protected:					\
  R<X,Y,Z >* rep;				\
public:						\
  T (R<X,Y,Z >* rep2);				\
  T (const T& x);				\
  ~T ();					\
  T& operator= (const T& x);			\
  R<X,Y,Z >* operator-> () const;

/******************************************************************************
* Indirect data structures with Boehm's GC
******************************************************************************/

#ifdef MML_SUPPORTS_GC // indirect structures for Boehm's GC

#define REP_STRUCT

#define INDIRECT_IMPL(T,R)						\
  inline T::T (R* rep2): rep (rep2) {}					\
  inline T::T (R* rep2, bool with_inc): rep (rep2) { (void) with_inc; }	\
  inline T::T (const T& x) : rep (x.rep) {}				\
  inline T::~T () {}							\
  inline T& T::operator= (const T& x) { rep= x.rep; return *this; }	\
  inline R* T::operator-> () const { return rep; }

#define INDIRECT_STATIC_IMPL(T,R)				\
  T::T (const T& x) : rep (x.rep) {}				\
  T::~T () {}							\
  T& T::operator= (const T& x) { rep= x.rep; return *this; }	\
  R* T::operator-> () const { return rep; }

#define INDIRECT_IMPL_1(T,R,XX,X)				\
  template<XX> T<X >::T (R<X >* rep2):				\
    rep (rep2) {}						\
  template<XX> T<X >::T (const T<X >& x):			\
    rep (x.rep) {}						\
  template<XX> T<X >::~T () {}					\
  template<XX> T<X >& T<X >::operator= (const T<X >& x) {	\
    rep= x.rep; return *this; }					\
  template<XX> R<X >* T<X >::operator-> () const {		\
    return rep; }

#define INDIRECT_IMPL_2(T,R,XX,X,YY,Y)					\
  template<XX,YY> T<X,Y >::T (R<X,Y >* rep2):				\
    rep (rep2) {}							\
  template<XX,YY> T<X,Y >::T (const T<X,Y >& x):			\
    rep (x.rep) {}							\
  template<XX,YY> T<X,Y >::~T () {}					\
  template<XX,YY> T<X,Y >& T<X,Y >::operator= (const T<X,Y >& x) {	\
    rep= x.rep; return *this; }						\
  template<XX,YY> R<X,Y >* T<X,Y >::operator-> () const {		\
    return rep; }

#define INDIRECT_IMPL_3(T,R,XX,X,YY,Y,ZZ,Z)				\
  template<XX,YY,ZZ> T<X,Y,Z >::T (R<X,Y,Z >* rep2):			\
    rep (rep2) {}							\
  template<XX,YY,ZZ> T<X,Y,Z >::T (const T<X,Y,Z >& x):			\
    rep (x.rep) {}							\
  template<XX,YY,ZZ> T<X,Y,Z >::~T () {}				\
  template<XX,YY,ZZ>							\
  T<X,Y,Z >& T<X,Y,Z >::operator= (const T<X,Y,Z >& x) {		\
    rep= x.rep; return *this; }						\
  template<XX,YY,ZZ> R<X,Y,Z >* T<X,Y,Z >::operator-> () const {	\
    return rep; }

/******************************************************************************
* Indirect data structures with reference counting
******************************************************************************/

#else

#define REP_STRUCT :public rep_struct

#define INDIRECT_IMPL(T,R)					\
  inline T::T (R* rep2): rep(rep2) {}				\
  inline T::T (R* rep2, bool with_inc): rep(rep2) {		\
    (void) with_inc; INC_COUNT (rep); }				\
  inline T::T (const T& x): rep(x.rep) { INC_COUNT (rep); }	\
  inline T::~T () { DEC_COUNT (rep); }				\
  inline R* T::operator -> () const { return rep; }		\
  inline T& T::operator = (const T& x) {			\
    INC_COUNT (x.rep); DEC_COUNT (rep);				\
    rep=x.rep; return *this; }

#define INDIRECT_STATIC_IMPL(T,R)			\
  T::T (const T& x): rep(x.rep) { INC_COUNT (rep); }	\
  T::~T () { DEC_COUNT (rep); }				\
  R* T::operator -> () const { return rep; }		\
  T& T::operator = (const T& x) {			\
    INC_COUNT (x.rep); DEC_COUNT (rep);			\
    rep=x.rep; return *this; }

#define INDIRECT_IMPL_1(T,R,XX,X)			\
  template<XX> T<X >::T (R<X >* rep2): rep(rep2) {}	\
  template<XX> T<X >::T (const T<X >& x): rep(x.rep) {	\
    INC_COUNT (rep); }					\
  template<XX> T<X >::~T () { DEC_COUNT (rep); }	\
  template<XX> R<X >* T<X >::operator -> () const {	\
    return rep; }					\
  template<XX> T<X >&					\
  T<X >::operator = (const T<X >& x) {			\
    INC_COUNT (x.rep); DEC_COUNT (rep);			\
    rep=x.rep; return *this; }

#define INDIRECT_IMPL_2(T,R,XX,X,YY,Y)				\
  template<XX,YY> T<X,Y >::T (R<X,Y >* rep2): rep(rep2) {}	\
  template<XX,YY> T<X,Y >::T (const T<X,Y >& x): rep(x.rep) {	\
    INC_COUNT (rep); }						\
  template<XX,YY> T<X,Y >::~T () { DEC_COUNT (rep); }		\
  template<XX,YY> R<X,Y >* T<X,Y >::operator -> () const {	\
    return rep; }						\
  template<XX,YY> T<X,Y >&					\
  T<X,Y >::operator = (const T<X,Y >& x) {			\
    INC_COUNT (x.rep); DEC_COUNT (rep);				\
    rep=x.rep; return *this; }

#define INDIRECT_IMPL_3(T,R,XX,X,YY,Y,ZZ,Z)				\
  template<XX,YY,ZZ> T<X,Y,Z >::T (R<X,Y,Z >* rep2): rep(rep2) {}	\
  template<XX,YY,ZZ> T<X,Y,Z >::T (const T<X,Y,Z >& x):			\
    rep(x.rep) { INC_COUNT (rep); }					\
  template<XX,YY,ZZ> T<X,Y,Z >::~T () { DEC_COUNT (rep); }		\
  template<XX,YY,ZZ> R<X,Y,Z >* T<X,Y,Z >::operator -> () const {	\
    return rep; }							\
  template<XX,YY,ZZ> T<X,Y,Z >&						\
  T<X,Y,Z >::operator = (const T<X,Y,Z >& x) {				\
    INC_COUNT (x.rep); DEC_COUNT (rep);					\
    rep=x.rep; return *this; }

#endif

} // namespace mml
#include "content.hpp"
#include "mmlout.hpp"
#include "defaults.hpp"
#include "variant.hpp"
#endif // __MML_BASE_HPP
