
/******************************************************************************
* MODULE     : integer.hpp
* DESCRIPTION: Integers
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_INTEGER_HPP
#define __MML_INTEGER_HPP
#include "mml_gmp.hpp"
namespace mml {

class integer;
class integer_rep;

class integer_rep REP_STRUCT {
private:
  mpz_t x;
public:
  inline integer_rep () { mpz_init (x); }
  inline ~integer_rep () { mpz_clear (x); }
  friend class integer;
  friend class rational;
};

class integer {
INDIRECT_PROTO (integer, integer_rep)
private:
  inline mpz_t& operator * () const { return rep->x; }

public:
  // constructors
  inline integer (): rep (new integer_rep ()) {}
  inline integer (signed int i):
    rep (new integer_rep ()) { mpz_set_si (rep->x, i); }
  inline integer (unsigned int i):
    rep (new integer_rep ()) { mpz_set_ui (rep->x, i); }
  inline integer (signed short int i):
    rep (new integer_rep ()) { mpz_set_si (rep->x, i); }
  inline integer (unsigned short int i):
    rep (new integer_rep ()) { mpz_set_ui (rep->x, i); }
  inline integer (signed long int i):
    rep (new integer_rep ()) { mpz_set_si (rep->x, i); }
  inline integer (unsigned long int i):
    rep (new integer_rep ()) { mpz_set_ui (rep->x, i); }
  inline integer (char* s, int base= 0):
    rep (new integer_rep ()) { mpz_set_str (rep->x, s, base); }

  // conversion
  inline friend int as_int (const integer& x) {
    return (int) mpz_get_si (*x); }

  // input/output
  inline friend ostream& operator << (ostream& out, const integer& x) {
    return print (out, *x); /*return out << (*x);*/ }

  // basic arithmetic
  inline friend integer operator - (const integer& x1) {
    integer r; mpz_neg (*r, *x1); return r; }
  inline friend integer operator + (const integer& x1, const integer& x2) {
    integer r; mpz_add (*r, *x1, *x2); return r; }
  inline friend integer operator - (const integer& x1, const integer& x2) {
    integer r; mpz_sub (*r, *x1, *x2); return r; }
  inline friend integer operator * (const integer& x1, const integer& x2) {
    integer r; mpz_mul (*r, *x1, *x2); return r; }
  inline friend integer operator / (const integer& x1, const integer& x2) {
    integer r; mpz_fdiv_q (*r, *x1, *x2); return r; }
  inline friend integer operator % (const integer& x1, const integer& x2) {
    integer r; mpz_fdiv_r (*r, *x1, *x2); return r; }
  inline friend integer operator << (const integer& x1, long int shift) {
    integer r; mpz_mul_2si (*r, *x1, shift); return r; }
  inline friend integer operator >> (const integer& x1, long int shift) {
    integer r; mpz_mul_2si (*r, *x1, -shift); return r; }
  inline friend integer& operator <<= (integer& x1, long int shift) {
    return x1= x1 << shift; }
  inline friend integer& operator >>= (integer& x1, long int shift) {
    return x1= x1 >> shift; }
  inline friend integer sqrt (const integer& x1) {
    integer r; mpz_sqrt (*r, *x1); return r; }
  inline friend integer pow (const integer& x1, const int& i) {
    assert (i >= 0);
    integer r; mpz_pow_ui (*r, *x1, (nat) i); return r; }
  inline friend integer pow (const integer& x1, const nat& i) {
    integer r; mpz_pow_ui (*r, *x1, i); return r; }
  inline friend integer abs (const integer& x1) {
    integer r; mpz_abs (*r, *x1); return r; }

  // comparing integer numbers
  inline friend bool operator == (const integer& x1, const integer& x2) {
    return mpz_cmp (*x1, *x2) == 0; }
  inline friend bool operator != (const integer& x1, const integer& x2) {
    return mpz_cmp (*x1, *x2) != 0; }
  inline friend bool operator < (const integer& x1, const integer& x2) {
    return mpz_cmp (*x1, *x2) < 0; }
  inline friend bool operator > (const integer& x1, const integer& x2) {
    return mpz_cmp (*x1, *x2) > 0; }
  inline friend bool operator <= (const integer& x1, const integer& x2) {
    return mpz_cmp (*x1, *x2) <= 0; }
  inline friend bool operator >= (const integer& x1, const integer& x2) {
    return mpz_cmp (*x1, *x2) >= 0; }
  inline friend int sign (const integer& x1) {
    int r= mpz_sgn (*x1); return (r<0? -1: (r==0? 0: 1)); }
  inline friend integer min (const integer& x1, const integer& x2) {
    return x1 <= x2? x1: x2; }
  inline friend integer max (const integer& x1, const integer& x2) {
    return x1 >= x2? x1: x2; }

  // miscellaneous functions
  inline friend nat hash (const integer& x) {
    __mpz_struct &rep= (*x)[0];
    if (rep._mp_size == 0) return 0;
    return ((nat) (rep._mp_d[0])) ^ ((nat) (rep._mp_size << 11));
  }

  friend class rational;
  inline friend integer num (const rational& x1);
  inline friend integer den (const rational& x1);
};
INDIRECT_IMPL (integer, integer_rep)

}
#endif // __MML_INTEGER_HPP
