
/******************************************************************************
* MODULE     : dense_polynomial.hpp
* DESCRIPTION: Dense univariate polynomials
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_DENSE_POLYNOMIAL_HPP
#define __MML_DENSE_POLYNOMIAL_HPP
#include "polynomial.hpp"
#include "jet.hpp"
namespace mml {
#define TMPL template<typename C, typename V>
#define Polynomial polynomial<C,dense<V> >
#define Polynomial_rep polynomial_rep<C,dense<V> >
TMPL int deg (const Polynomial& P);
TMPL nat dim (const Polynomial& P);
TMPL C*  seg (const Polynomial& P);

/******************************************************************************
* Dense polynomials
******************************************************************************/

template<typename C, typename V>
class Polynomial_rep REP_STRUCT {
  C*  a; // size of a can be larger than n due to normalization
  nat n;

public:
  void normalize () { while ((n != 0) && (a[n-1] == 0)) n--; }
  polynomial_rep (C* a2, nat n2): a (a2), n (n2) { normalize (); }
  ~polynomial_rep () { delete[] a; }
  void extend (nat l) {
    C* b= new C[l];
    jet_copy (b, a, n);
    jet_clear (b+n, l-n);
    delete[] a;
    a= b;
    n= l;
  }

  friend class Polynomial;
  friend int deg LESSGTR (const Polynomial& P);
  friend nat dim LESSGTR (const Polynomial& P);
  friend C*  seg LESSGTR (const Polynomial& P);
};

template<typename C, typename V>
class Polynomial {
INDIRECT_PROTO_2 (polynomial, polynomial_rep, C, dense<V>)
public:
  polynomial () {
    rep= new Polynomial_rep (new C[0], 0);
  }
  polynomial (int x) {
    C* a= new C[1];
    a[0]= x;
    rep= new Polynomial_rep (a, 1);
  }
  polynomial (const C& x, nat n=0) {
    C* a= new C[n+1];
    jet_clear (a, n);
    a[n]= x;
    rep= new Polynomial_rep (a, n+1);
  }
  polynomial (C* a, nat n) {
    rep= new Polynomial_rep (a, n);
  }
  polynomial (const iterator<C>& it) {
    nat i, l=0;
    rep= new Polynomial_rep (new C[0], 0);
    for (i=0; busy (it); i++, ++it) {
      if (i >= l) {
	l= max ((nat) 1, l<<1);
	rep->extend (l);
      }
      rep->a[i]= *it;
    }
    rep->normalize ();
  }
  C operator[] (nat i) const { return rep->a[i]; }
};
INDIRECT_IMPL_2 (polynomial, polynomial_rep,
		 typename C, C, typename V, dense<V>)

TMPL nat dim (const Polynomial& P) { return P->n; }
TMPL int deg (const Polynomial& P) { return ((int) P->n) - 1; }
TMPL C*  seg (const Polynomial& P) { return P->a; }

/******************************************************************************
* Basic arithmetic operations
******************************************************************************/

TMPL Polynomial
operator + (const Polynomial& P1, const Polynomial& P2) {
  nat m= min (dim (P1), dim (P2));
  nat n= max (dim (P1), dim (P2));
  C* r= new C[n];
  jet_add<C> (r, seg (P1), seg (P2), m);
  if (dim (P1) > m) jet_copy<C> (r+m, seg(P1)+m, n-m);
  if (dim (P2) > m) jet_copy<C> (r+m, seg(P2)+m, n-m);
  return Polynomial (r, n);
}

TMPL Polynomial
operator - (const Polynomial& P1, const Polynomial& P2) {
  nat m= min (dim (P1), dim (P2));
  nat n= max (dim (P1), dim (P2));
  C* r= new C[n];
  jet_sub<C> (r, seg (P1), seg (P2), m);
  if (dim (P1) > m) jet_copy<C> (r+m, seg(P1)+m, n-m);
  if (dim (P2) > m) jet_neg<C> (r+m, seg(P2)+m, n-m);
  return Polynomial (r, n);
}

TMPL Polynomial
operator - (const Polynomial& P) {
  nat n= dim (P);
  C* r= new C[n];
  jet_neg<C> (r, seg (P), n);
  return Polynomial (r, n);
}

TMPL Polynomial
operator * (const Polynomial& P1, const Polynomial& P2) {
  nat n1= dim (P1), n2= dim (P2);
  if (n1 == 0 || n2 == 0) return Polynomial ();
  C* r= new C[n1+n2-1];
  jet_mul<C,V> (r, seg (P1), seg (P2), n1, n2);
  return Polynomial (r, n1+n2-1);
}

TMPL Polynomial
operator * (const Polynomial& P, const C& c) {
  nat n= dim (P);
  C* r= new C[n];
  jet_mul<C> (r, seg (P), c, n);
  return Polynomial (r, n);
}

TMPL Polynomial
operator * (const C& c, const Polynomial& P) {
  nat n= dim (P);
  C* r= new C[n];
  jet_mul<C> (r, seg (P), c, n);
  return Polynomial (r, n);
}

TMPL Polynomial
operator / (const Polynomial& P, const C& c) {
  nat n= dim (P);
  C* r= new C[n];
  jet_div<C> (r, seg (P), c, n);
  return Polynomial (r, n);
}

TMPL Polynomial&
operator += (Polynomial& P1, const Polynomial& P2) {
  nat n1= dim (P1), n2= dim (P2);
  if (n1 < n2) P1->extend (n2);
  jet_add<C> (seg (P1), seg (P2), n2);
  if (n1 == n2) P1->normalize ();
  return P1;
}

TMPL Polynomial&
operator -= (Polynomial& P1, const Polynomial& P2) {
  nat n1= dim (P1), n2= dim (P2);
  if (n1 < n2) P1->extend (n2);
  jet_sub<C> (seg (P1), seg (P2), n2);
  if (n1 == n2) P1->normalize ();
  return P1;
}

TMPL Polynomial&
operator *= (Polynomial& P, const C& c) {
  jet_mul<C> (seg (P), c, dim (P));
  P->normalize ();
  return P;
}

TMPL Polynomial&
operator /= (Polynomial& P, const C& c) {
  jet_div<C> (seg (P), c, dim (P));
  P->normalize ();
  return P;
}

/******************************************************************************
* Comparisons
******************************************************************************/

TMPL int
sign (const Polynomial& P) {
  nat n= dim (P);
  if (n == 0) return 0;
  return sign (P[n-1]);
}

TMPL int
compare (const Polynomial& P1, const Polynomial& P2) {
  nat n1= dim (P1), n2= dim (P2);
  if (n1 > n2) return sign (P1[n1-1]);
  if (n2 > n1) return -sign (P2[n2-1]);
  return jet_cmp (seg (P1), seg (P2), n1);
}

/******************************************************************************
* Syntactic sugar
******************************************************************************/

ARITH_INT_SUGAR(TMPL,Polynomial,C);
SHIFT_INT_SUGAR(TMPL,Polynomial,C);
COMPARE_SUGAR(TMPL,Polynomial);

#undef TMPL
#undef Polynomial
#undef Polynomial_rep
} // namespace mml
#endif // __MML_DENSE_POLYNOMIAL_HPP
