
/******************************************************************************
* MODULE     : distribution.hpp
* DESCRIPTION: Distributions
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_DISTRIBUTION_HPP
#define __MML_DISTRIBUTION_HPP
#include "string.hpp"
#include "dense_vector.hpp"
namespace mml {
#define TMPL_DEF template<typename C,typename R=double>
#define TMPL template<typename C, typename R>
#define Distribution distribution<C,R>
#define Distribution_rep distribution_rep<C,R>
TMPL class distribution;

/******************************************************************************
* The abstract distribution class
******************************************************************************/

TMPL_DEF
class distribution_rep REP_STRUCT {
public:
  virtual C pick (const vector<nat>& n) = 0;
  // Generate a random element in the distribution of size n.
  // The size may be an empty vector (when sizes are ignored),
  // simple positive integers (giving rise to univariate generating functions),
  // or more general vectors (for specifying further parameters and giving
  // rise to multivariate generating functions.

  virtual R count (const vector<nat>& n) {
    if (dim(n) == 0 || (dim(n) == 1 && n[0] == 1)) return 1;
    return 0;
  }
  // Returns the "number" of objects of a given size.
  // More precisely, the probability that an object of size (n1,...nk)
  // is chosen equals count(n1,...,nk) / sum_{m1,...,mk} count(m1,...,mk).
  // Notice that, count (n1,...,n(k-1))= sum_mk count(n1,...,n(k-1),mk).
  // In other words, if less size parameters are specifies,
  // then the remaining ones may be chosen freely.

  virtual content expression () {
    return content ("distribution"); }

public:
  distribution_rep () {}
  virtual ~distribution_rep () {}
  friend class Distribution;
};

TMPL_DEF
class distribution {
INDIRECT_PROTO_2 (distribution, distribution_rep, C, R)
public:
  C operator() () const { return rep->pick (vec<nat> ()); }
  C operator() (nat n) const { return rep->pick (vec<nat> (n)); }
  C operator() (nat n1, nat n2) const { return rep->pick (vec<nat> (n1, n2)); }
  C operator() (const vector<nat>& v) const { return rep->pick (v); }
};
INDIRECT_IMPL_2 (distribution, distribution_rep, typename C, C, typename R, R)

TMPL content as_content (const Distribution& db) {
  return db->expression (); }
PRINT_SUGAR(TMPL,Distribution);

/******************************************************************************
* Uniform distributions
******************************************************************************/

template<typename C> C
unif_random () {
  return (C) (((double) rand()) / RAND_MAX);
}

template<typename C> C
unif_random (C right) {
  return (C) (((double) right) * ((double) rand()) / RAND_MAX);
}

template<typename C> C
unif_random (C left, C right) {
  return left + (C) (((double) (right-left)) * ((double) rand()) / RAND_MAX);
}

TMPL_DEF
class uniform_distribution_rep: public Distribution_rep {
  C left, right;

public:
  virtual C pick (const vector<nat>& n) {
    (void) n;
    assert (dim(n) == 0 || n[0] == 1);
    return unif_random<C> (left, right); }
  virtual content expression () {
    return as_content ("uniform", left, right); }

public:
  uniform_distribution_rep (const C& left2, const C& right2):
    left (left2), right (right2) {}
};

TMPL Distribution
uniform () {
  return Distribution (new uniform_distribution_rep<C,R> (0, 1));
}

TMPL Distribution
uniform (const C& r) {
  return Distribution (new uniform_distribution_rep<C,R> (0, r));
}

TMPL Distribution
uniform (const C& l, const C& r) {
  return Distribution (new uniform_distribution_rep<C,R> (l, r));
}

template<typename C> distribution<C>
uniform () {
  return distribution<C> (new uniform_distribution_rep<C> (0, 1));
}

template<typename C> distribution<C>
uniform (const C& r) {
  return distribution<C> (new uniform_distribution_rep<C> (0, r));
}

template<typename C> distribution<C>
uniform (const C& l, const C& r) {
  return distribution<C> (new uniform_distribution_rep<C> (l, r));
}

/******************************************************************************
* Adding distributions
******************************************************************************/

TMPL_DEF
class sum_distribution_rep: public Distribution_rep {
  Distribution db1, db2;

public:
  C pick (const vector<nat>& n) {
    double c1= (double) db1->count (n);
    double c2= (double) db2->count (n);
    if (unif_random<double> () < c1/(c1+c2)) return db1 (n);
    else return db2 (n); }
  R count (const vector<nat>& n) {
    return db1->count (n) + db2->count (n); }
  content expression () {
    return as_content ("+", db1, db2); }

public:
  sum_distribution_rep (const Distribution& db1b, const Distribution& db2b):
    db1(db1b), db2(db2b) {}
};

TMPL Distribution
operator + (const Distribution& db1, const Distribution& db2) {
  return Distribution (new sum_distribution_rep<C> (db1, db2));
}

/******************************************************************************
* Specializing distributions
******************************************************************************/

static vector<nat>
merge (const vector<nat>& v1, vector<nat>& v2) {
  vector<nat> r (dim (v1) + dim (v2));
  for (nat i=0; i<dim(v1); i++) r[i]= v1[i];
  for (nat i=0; i<dim(v2); i++) r[i+dim(v1)]= v2[i];
  return r;
}

TMPL_DEF
class specialize_distribution_rep: public Distribution_rep {
  Distribution db;
  vector<nat>  v;

public:
  C pick (const vector<nat>& n) {
    return db (merge (n, v)); }
  R count (const vector<nat>& n) {
    return db->count (merge (n, v)); }
  content expression (ostream& out) {
    return as_content ("specialize", db, v); }

public:
  specialize_distribution_rep (const Distribution& db2, const vector<nat>& v2):
    db(db2), v(v2) {}
};

TMPL Distribution
specialize (const Distribution& db, nat n) {
  return Distribution (new specialize_distribution_rep<C> (db, vec(n)));
}

TMPL Distribution
specialize (const Distribution& db, const vector<nat>& v) {
  return Distribution (new specialize_distribution_rep<C> (db, v));
}

#undef TMPL_DEF
#undef TMPL
#undef Distribution
#undef Distribution_rep
} // namespace mml
#endif // __MML_DISTRIBUTION_HPP
