
/******************************************************************************
* MODULE     : series.hpp
* DESCRIPTION: Relaxed series
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_SERIES_HPP
#define __MML_SERIES_HPP
#include "dense_polynomial.hpp"
namespace mml {
#define TMPL_DEF template<typename C, typename V=naive>
#define TMPL template<typename C, typename V>
#define Series series<C,V>
#define Series_rep series_rep<C,V>
TMPL class series;

/******************************************************************************
* Dense series
******************************************************************************/

TMPL_DEF
class series_rep REP_STRUCT {
protected:
  C*  a;  // coefficients
  nat n;  // number of computed coefficients
  nat l;  // number of allocated coefficients
          // a[n],...,a[l-1] may be used by relaxed computations

  inline series_rep (): a (new C[0]), n (0), l (0) {}
  inline virtual ~series_rep () { delete[] a; }
  virtual C next () = 0;

public:
  virtual content expression () = 0;
  virtual void set_precision (nat l2);
  virtual void increase_precision (nat l2=0);
  virtual C& initial (nat n);
  virtual Series& equation ();
  friend class Series;
};

TMPL_DEF
class series {
INDIRECT_PROTO_2 (series, series_rep, C, V)
public:
  series (int c= 0);
  series (const C& c, nat deg= 0);
  series (const polynomial<C>& P);
  series (const iterator<C>& it, const string& name= "explicit");
  series (C (*coeffs) (nat), const string& name= "explicit");
  C  operator [] (nat n) const;
  C& operator () (nat n) const;
  C* operator () (nat start, nat end) const;
};
INDIRECT_IMPL_2 (series, series_rep, typename C, C, typename V, V)

TMPL string var (const Series& f) { (void) f; return the_series_var; }

/******************************************************************************
* Routines on dense series
******************************************************************************/

TMPL void
Series_rep::set_precision (nat l2) {
  if (l2 <= l) return;
  C* b= new C[l2];
  jet_copy (b, a, l);
  jet_clear (b+l, l2-l);
  delete[] a;
  a= b;
  l= l2;
}

TMPL void
Series_rep::increase_precision (nat l2) {
  l2= max (l2, l);
  set_precision (max ((nat) 1, l2 << 1));
}

TMPL C&
Series_rep::initial (nat i) {
  fatal_error ("Not an implicit series",
	       "series_rep<C,V>::initial",
	       "series.hpp");
  return a[0];
}

TMPL Series&
Series_rep::equation () {
  fatal_error ("Not an implicit series",
	       "series_rep<C,V>::equation",
	       "series.hpp");
  return *((Series*) ((void*) this));
}

TMPL C
Series::operator [] (nat n) const {
  if (n <  rep->n) return rep->a[n];
  if (n >= rep->l) rep->increase_precision (n);
  while (rep->n <= n) {
    rep->a[rep->n]= rep->next ();
    rep->n++;
  }
  // cout << (*this) << " [" << n << "] -> " << rep->a[n] << "\n";
  return rep->a[n];
}

TMPL C*
Series::operator () (nat start, nat end) const {
  if (end <= rep->n) return rep->a + start;
  if (end >= rep->l) rep->increase_precision (end);
  while (rep->n < end) {
    rep->a[rep->n]= rep->next ();
    rep->n++;
  }
  return rep->a + start;
}

TMPL C&
Series::operator () (nat n) const {
  return rep->initial (n);
}

/******************************************************************************
* Series iterator
******************************************************************************/

TMPL_DEF
class series_iterator_rep: public iterator_rep<C> {
  Series f;  // the series to traverse
  nat i;     // current position
protected:
  bool is_busy () { return true; }
  void advance () { i++; }
  C current () { return f[i]; }
public:
  series_iterator_rep (const Series& f2): f(f2), i(0) {}
};

TMPL iterator<C>
iterate (const Series& f) {
  return iterator<C> (new series_iterator_rep<C,V> (f));
}

/******************************************************************************
* Printing series
******************************************************************************/

TMPL content
as_content (const Series& f) {
  if (the_series_expr) return f->expression ();
  content s= 0;
  for (nat i=0; i<the_series_prec; i++)
    s= s + as_content (f[i]) * pow (as_content (var (f)), content (i));
  s= s + apply ("O", pow (as_content (var (f)), content (the_series_prec)));
  return s;
}

PRINT_SUGAR(TMPL,Series)

/******************************************************************************
* Explicit series
******************************************************************************/

TMPL_DEF
class polynomial_series_rep: public Series_rep {
  polynomial<C> P;
public:
  polynomial_series_rep (const polynomial<C>& P2): P (P2) {}
  C next () { return this->n < dim(P)? P[this->n]: 0; }
  content expression () { return as_content (P); }
};

TMPL
Series::series (int c) {
  rep= new polynomial_series_rep<C,V> (polynomial<C> (C(c), 0));
}

TMPL
Series::series (const C& c, nat deg) {
  rep= new polynomial_series_rep<C,V> (polynomial<C> (c, deg));
}

TMPL
Series::series (const polynomial<C>& P) {
  rep= new polynomial_series_rep<C,V> (P);
}

TMPL_DEF
class func_series_rep : public Series_rep {
  C (*coeffs) (nat);
  string name;
public:
  func_series_rep (C (*coeffs2) (nat), const string& name2):
    coeffs (coeffs2), name (name2) {}
  C next () { return coeffs (this->n); }
  content expression () { return as_content (name); }
};

TMPL
Series::series (C (*coeffs) (nat), const string& name) {
  rep= new func_series_rep<C> (coeffs, name);
}

TMPL_DEF
class iterator_series_rep : public Series_rep {
  iterator<C> it;
  string name;
public:
  iterator_series_rep (const iterator<C>& it2, const string& name2):
    it (it2), name (name2) {}
  C next () {
    if (busy (it)) {
      C c= *it; ++it; return c; }
    else return C(0); }
  content expression () { return as_content (name); }
};

TMPL
Series::series (const iterator<C>& it, const string& name) {
  rep= new iterator_series_rep<C> (it, name);
}

/******************************************************************************
* Sums and products
******************************************************************************/

TMPL_DEF
class sum_series_rep : public Series_rep {
  Series f, g;
public:
  sum_series_rep (const Series& f2, const Series& g2):
    f(f2), g(g2) {}
  C next () { return f[this->n] + g[this->n]; }
  content expression () { return as_content (f) + as_content (g); }
};

TMPL Series
operator + (const Series& f, const Series& g) {
  return Series (new sum_series_rep<C,V> (f, g));
}

TMPL_DEF
class diff_series_rep : public Series_rep {
  Series f, g;
public:
  diff_series_rep (const Series& f2, const Series& g2):
    f(f2), g(g2) {}
  C next () { return f[this->n] - g[this->n]; }
  content expression () { return as_content (f) - as_content (g); }
};

TMPL Series
operator - (const Series& f) {
  return Series (new diff_series_rep<C,V> (0, f));
}

TMPL Series
operator - (const Series& f, const Series& g) {
  return Series (new diff_series_rep<C,V> (f, g));
}

/******************************************************************************
* Scalar multiplication and division
******************************************************************************/

TMPL_DEF
class scprod_series_rep : public Series_rep {
  Series f;
  C c;
public:
  scprod_series_rep (const Series& f2, C c2): f(f2), c(c2) {}
  C next () { return f[this->n] * c; }
  content expression () { return as_content (c) * as_content (f); }
};

TMPL Series
operator * (const Series& f, const C& c) {
  return Series (new scprod_series_rep<C,V> (f, c));
}

TMPL Series
operator * (const C& c, const Series& f) {
  return Series (new scprod_series_rep<C,V> (f, c));
}

TMPL_DEF
class scdiv_series_rep : public Series_rep {
  Series f;
  C c;
public:
  scdiv_series_rep (const Series& f2, C c2): f(f2), c(c2) {}
  C next () { return f[this->n] / c; }
  content expression () { return as_content (f) / as_content (c); }
};

TMPL Series
operator / (const Series& f, const C& c) {
  return Series (new scdiv_series_rep<C,V> (f, c));
}

/******************************************************************************
* Multiplication
******************************************************************************/

TMPL_DEF
class prod_series_rep : public Series_rep {
  Series f, g;
public:
  prod_series_rep (const Series& f2, const Series& g2):
    f(f2), g(g2) {}
  C next () {
    nat i;
    C acc= f[0] * g[this->n];
    for (i=1; i<=this->n; i++)
      acc += f[i] * g[this->n-i];
    return acc;
  }
  content expression () { return as_content (f) * as_content (g); }
};

TMPL Series
operator * (const Series& f, const Series& g) {
  return Series (new prod_series_rep<C,V> (f, g));
}

/******************************************************************************
* Other linear time operations
******************************************************************************/

TMPL_DEF
class shift_series_rep : public Series_rep {
  Series f;
  int shift;
public:
  shift_series_rep (const Series& f2, int shift2):
    f (f2), shift (shift2) {}
  C next () { return shift > ((int) this->n)? 0: f[this->n - shift]; }
  content expression () {
    if (shift>0) return as_content ("<<", f, shift);
    else return as_content (">>", f, -shift); }
};

TMPL Series
operator << (const Series& f, int shift) {
  return Series (new shift_series_rep<C,V> (f, shift));
}

TMPL Series
operator >> (const Series& f, int shift) {
  return Series (new shift_series_rep<C,V> (f, -shift));
}

TMPL_DEF
class derive_series_rep : public Series_rep {
  Series f;
public:
  derive_series_rep (const Series& f2): f (f2) {}
  C next () { return (this->n + 1) * f[this->n + 1]; }
  content expression () { return as_content ("derive", f); }
};

TMPL Series
derive (const Series& f) {
  return Series (new derive_series_rep<C,V> (f));
}

TMPL_DEF
class integrate_series_rep : public Series_rep {
  Series f;
public:
  integrate_series_rep (const Series& f2): f (f2) {}
  C next () { return this->n == 0? 0: f[this->n - 1] / this->n; }
  content expression () { return as_content ("integrate", f); }
};

TMPL Series
integrate (const Series& f) {
  return Series (new integrate_series_rep<C,V> (f));
}

TMPL_DEF
class dilated_series_rep : public Series_rep {
  Series f;
  nat p;
public:
  dilated_series_rep (const Series& f2, int p2): f (f2), p (p2) {}
  C next () { return this->n % p == 0? f[this->n / p]: 0; }
  content expression () { return as_content ("dilate", f, p); }
};

TMPL Series
dilate (const Series& f, nat p) {
  return Series (new dilated_series_rep<C,V> (f, p));
}

/******************************************************************************
* Comparisons
******************************************************************************/

TMPL int
sign (const Series& f) {
  for (nat n=0; n<the_series_zt; n++) {
    int sgn= sign (f[n]);
    if (sgn != 0) return sgn;
  }
  return 0;
}

TMPL int
compare (const Series& f1, const Series& f2) {
  for (nat n=0; n<the_series_zt; n++) {
    int sgn= compare (f1[n], f2[n]);
    if (sgn != 0) return sgn;
  }
  return 0;
}

/******************************************************************************
* Syntactic sugar
******************************************************************************/

ARITH_INT_SUGAR(TMPL,Series,C);
SHIFT_INT_SUGAR(TMPL,Series,C);
COMPARE_SUGAR(TMPL,Series);

#undef TMPL_DEF
#undef TMPL
#undef Series
#undef Series_rep
} // namespace mml
#include "implicit_series.hpp"
#endif // __MML_SERIES_HPP
