
/******************************************************************************
* MODULE     : symbolic.hpp
* DESCRIPTION: Symbolic expressions
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __SYMBOLIC_HPP
#define __SYMBOLIC_HPP
#include "string.hpp"
#include "dense_vector.hpp"
#include "symbolic_defs.hpp"
namespace mml {

class symbolic;
class symbolic_aux;

class symbolic_rep: public rep_struct {
public: // FIXME: should become protected
  // accessors and syntactic identity
  virtual symbolic  type ();
  virtual nat       priority ();
  virtual bool      is_atomic ();
  virtual string    leaf ();
  virtual nat       arity ();
  virtual symbolic  child (nat i);
  virtual nat       hash ();
  virtual bool      identical (symbolic_rep* rep);

  // arithmetic operations
  virtual symbolic neg (const symbolic& s1);
  virtual symbolic add (const symbolic& s1, const symbolic& s2);
  virtual symbolic sub (const symbolic& s1, const symbolic& s2);
  virtual symbolic mul (const symbolic& s1, const symbolic& s2);
  virtual symbolic div (const symbolic& s1, const symbolic& s2);
  virtual symbolic pow (const symbolic& s1, const symbolic& s2);

  // properties
  virtual bool is_abelian ();      // abelian group for plus
  virtual bool is_associative ();  // associative multiplication
  virtual bool is_commutative ();  // commutative multiplication
  virtual bool is_distributive (); // distribute sums when multiplying
  virtual bool is_scalar ();       // treat as scalar for linear algebra

public:
  inline symbolic_rep () {}
  inline virtual ~symbolic_rep () {}
  friend class symbolic;
};

class symbolic {
protected:
  symbolic_rep* rep;

public:
  inline symbolic (const symbolic& x);
  inline symbolic_rep* operator -> () const;
  symbolic (symbolic_rep* rep);
  symbolic& operator = (const symbolic& x);
  ~symbolic ();

public:
  symbolic ();
  symbolic (int i);
  symbolic (const integer& i);
  symbolic (const rational& r);
  symbolic (char* s);
  symbolic (const string& s);

  inline friend bool is_atomic (const symbolic& s);
  inline friend string L (const symbolic& s);
  friend integer as_integer  (const symbolic& s);
  friend rational as_rational (const symbolic& s);
  inline friend nat N (const symbolic& s);
  inline symbolic operator [] (nat i) const;
  inline friend nat hash (const symbolic& s);
  inline bool operator == (const symbolic& s) const;
  inline bool operator != (const symbolic& s) const;
  inline bool operator <  (const symbolic& s) const ;
  inline bool operator <= (const symbolic& s) const;
  inline bool operator >  (const symbolic& s) const;
  inline bool operator >= (const symbolic& s) const;
  friend ostream& operator << (ostream& out, const symbolic& s);

  friend class symbolic_rep;
  friend class symbolic_atom_rep;
  friend class symbolic_integer_rep;
  friend class symbolic_compound_rep;
};

inline symbolic::symbolic (const symbolic& x): rep(x.rep) { INC_COUNT (rep); }
inline symbolic_rep* symbolic::operator -> () const { return rep; }
inline bool is_atomic (const symbolic& s) { return s->is_atomic (); }
inline string L (const symbolic& s) { return s->leaf (); }
inline nat N (const symbolic& s) { return s->arity (); }
inline symbolic symbolic::operator [] (nat i) const { return rep->child (i); }
inline nat hash (const symbolic& s) { return ((nat) s.rep) >> 2; }

inline bool symbolic::operator == (const symbolic& s) const {
  return rep == s.rep; }
inline bool symbolic::operator != (const symbolic& s) const {
  return rep != s.rep; }
inline bool symbolic::operator <  (const symbolic& s) const {
  return rep <  s.rep; }
inline bool symbolic::operator <= (const symbolic& s) const {
  return rep <= s.rep; }
inline bool symbolic::operator >  (const symbolic& s) const {
  return rep >  s.rep; }
inline bool symbolic::operator >= (const symbolic& s) const {
  return rep >= s.rep; }

symbolic symbolic_compound (const vector<symbolic>& fun_args);
symbolic flatten (const symbolic& s);
void lisp_print (ostream& out, const symbolic& s);
void pretty_print (ostream& out, const symbolic& s);

symbolic apply (const symbolic& fun);
symbolic apply (const symbolic& fun, const symbolic& s1);
symbolic apply (const symbolic& fun, const symbolic& s1, const symbolic& s2);
symbolic apply (const symbolic& fun, const vector<symbolic>& args);

symbolic operator - (const symbolic& s1);
symbolic operator + (const symbolic& s1, const symbolic& s2);
symbolic operator - (const symbolic& s1, const symbolic& s2);
symbolic operator * (const symbolic& s1, const symbolic& s2);
symbolic operator / (const symbolic& s1, const symbolic& s2);
symbolic operator % (const symbolic& s1, const symbolic& s2);
symbolic pow (const symbolic& s1, const symbolic& s2);

} // namespace mml
#endif // __SYMBOLIC_HPP
